﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;     // Timer

namespace LP_MineSweeper {
	public class Game {
		private readonly Random rg=new Random();
		private readonly int nRows;
		private readonly int nCols;
		private readonly int nMines;
		private Cell[,] board;

		private DateTime startTime;
		private Timer timer;
		private string status;
		private bool aborted;
		public bool revealAllMines;

		// delegates used to alter Controls on the user interface
		public Action<string> AddToLog;
		public Action<double> ShowTime;
		public Action<string> ShowStatus;
		public Action<int> ShowHiddenMineCount;
		public Action<string> PlaySound;

		public Game(Action<string> addToLog, int nRows, int nCols, int nMines) {
			this.AddToLog = addToLog;
			this.nRows = nRows;
			this.nCols = nCols;
			this.nMines=nMines;
			createAllCells();
			placeRandomMines();
			status = "Awaiting first click";
		}

		// log one line of text
		private void log(string text) {
			AddToLog?.Invoke(text);
		}

		public void Stop() {
			timer?.Stop();
		}

		private void createAllCells() {
			board = new Cell[nRows, nCols];
			for (int r = 0; r < nRows; r++) {
				for (int c = 0; c < nCols; c++) {
					Cell cell = new Cell(this, r, c);
					board[r, c] = cell;
				}
			}
		}
		// place mines at random and calculate neighbours
		// (every mine placed increments the danger level of all neighboring cells)
		private void placeRandomMines() {
			int todo = nMines;
			while (todo > 0) {
				int r = rg.Next(nRows);
				int c = rg.Next(nCols);
				Cell cell = board[r, c];
				if (!cell.IsMine) {
					cell.IsMine = true;
					todo--;
					log("createMines: " + cell);
					for (int rr = r - 1; rr <= r + 1; rr++) {
						if (rr >= 0 && rr < nRows) {
							for (int cc = c - 1; cc <= c + 1; cc++) {
								if (cc >= 0 && cc < nCols) {
									if (rr != r || cc != c) board[rr, cc].IncrementDangerLevel();
								}
							}
						}
					}
				}
			}
		}

		// the board isn't public, so stuffing all cells in the UI container 
		// will have to do be executed one cell at a time
		public Cell GetCell(int row, int col) {
			return board[row, col];
		}

		[System.Diagnostics.CodeAnalysis.SuppressMessage("Style", "IDE0017:Simplify object initialization", Justification = "Silly suggestion")]
		public void UpdateGame(Cell cell) {
			//log("updateStats" + cell?.ToString());
			// timer only starts running when a first cell gets clicked
			if (cell != null && timer==null) {
				startTime = DateTime.Now;
				timer = new Timer();
				timer.Interval = 50;
				timer.Tick += timerTick;
				timer.Start();
			}
			if (cell!=null && cell.IsMine) PlaySound?.Invoke("explosion");
			floodFill();
			int revealedCount = 0;
			int minesHitCount = 0;
			int flaggedCount = 0;
			for (int r = 0; r < nRows; r++) {
				for (int c = 0; c < nCols; c++) {
					cell = board[r, c];
					if (cell.IsRevealed) {
						revealedCount++;
						if (cell.IsMine) {
							minesHitCount++;
						}
					}
					if (cell.IsFlagged) {
						flaggedCount++;
					}
				}
			}
			ShowHiddenMineCount?.Invoke(nMines - flaggedCount);
			status = null;
			if (aborted) {
				status="Game aborted";
			} else if (minesHitCount != 0) {
				status="Bad luck";
			} else if (revealedCount + nMines == nRows * nCols) {
				// the game is won when all non-mine cells have been revealed,
				// there is no need to flag the mines (it could be a waste of time!)
				status="Congratulations";
			}
			if (status!=null) {
				timer?.Stop();
				ShowStatus?.Invoke(status);
			}
		}

		private void timerTick(object sender, EventArgs e) {
			TimeSpan elapsed = DateTime.Now - startTime;
			ShowTime?.Invoke(elapsed.TotalSeconds);
		}

		private void floodFill() {
			// we reveal all cells that have a safe and revealed neighbour
			// and we keep doing so until no thing is gained any more
			bool more = true;
			for (int iter = 1; more; iter++) {  // keep doing this until 'more' stays false
				log("floodFill iter=" + iter);
				more = false;
				// check all cells one by one from top to bottom, and left to right
				for (int r = 0; r < nRows; r++) {
					for (int c = 0; c < nCols; c++) {
						Cell cell = board[r, c];
						if (cell.IsKnownToBeSafe) {
							for (int rr = r - 1; rr <= r + 1; rr++) {
								if (rr >= 0 && rr < nRows) {
									for (int cc = c - 1; cc <= c + 1; cc++) {
										if (cc >= 0 && cc < nCols) {
											if (rr != r || cc != c) {
												Cell cell2 = board[rr, cc];
												if (!cell2.IsRevealed) {
													cell2.Reveal();
													more = true;
												}
											}
										}
									}
								}
							}

						}
					}
				}
			}
		}

		public void Abort(bool giveAwayAllMines) {
			log("Game.Abort");
			aborted = true;
			if (giveAwayAllMines) {
				// Invalidating the entire board is too slow, only hidden mines need a repaint
				for (int r = 0; r < nRows; r++) {
					for (int c = 0; c < nCols; c++) {
						Cell cell = board[r, c];
						if (cell.IsMine) cell.GiveAway();
					}
				}
				UpdateGame(null);
			}
		}
	}
}
